/*

MyPing - sample program

Keisuke Oyama

29-JUN-2005 - Refresh

*/

#include <winsock2.h>
#include <stdio.h>
#include "pingex.h"
#include "printutils.h"


#define ICMP_DATASIZE  (32)


//////////////////////////////////////////////////////////////////////////


USHORT CheckSum(USHORT *buffer, int size) {

    unsigned long cksum=0;

    while (size > 1) 
    {
        cksum += *buffer++;
        size -= sizeof(USHORT);
    }
    if (size) 
    {
        cksum += *(UCHAR*)buffer;
    }
    cksum = (cksum >> 16) + (cksum & 0xffff);
    cksum += (cksum >>16);
    return (USHORT)(~cksum);
}


//////////////////////////////////////////////////////////////////////////


BOOL BuildICMPEchoPacket(char* pBuff)
{	
	static char chData = 'a';
	PICMP_HDR pICMPHdr;
	char* pDataTop;
	
	pICMPHdr = (PICMP_HDR) pBuff;
	pICMPHdr->icmp_type = 8; // echo request.
	pICMPHdr->icmp_code = 0;
	pICMPHdr->icmp_id = (unsigned short) GetCurrentProcessId();
	pICMPHdr->icmp_checksum = 0; // zero field before computing checksum
	pICMPHdr->icmp_sequence = 0;
	
	pDataTop = pBuff;
	pDataTop += sizeof(ICMP_HDR);
	memset(pDataTop, chData++, ICMP_DATASIZE);

	pICMPHdr->icmp_timestamp = GetTickCount();

	pICMPHdr->icmp_checksum = 
		CheckSum((USHORT*) pICMPHdr, sizeof(ICMP_HDR) + ICMP_DATASIZE);

	return TRUE;
	
}


//////////////////////////////////////////////////////////////////////////


void ShowUsage()
{
	printf("\nMyPing network utility.\nCopyright(C) 2005 Keisuke Oyama <i96815@yahoo.co.jp>\n\n");
	printf("Usage: myping <target>\n");
	printf("  target    hostname or ip v4 address.\n");
}


//////////////////////////////////////////////////////////////////////////


BOOL ResolveAddress(
	char* pArg, SOCKADDR_IN* pAddr, char* pHostName, size_t nSize) 
{

	INT nRet;
	addrinfo hints, *outAddrInfo = NULL;
	
	ZeroMemory(&hints, sizeof(addrinfo));

	hints.ai_flags = AI_CANONNAME;
	hints.ai_family = PF_INET; // IPv4
	hints.ai_socktype = SOCK_STREAM;

	nRet = getaddrinfo( pArg, NULL, &hints, &outAddrInfo);

	if(0 != nRet ) {
		printf("%s\n", gai_strerror(nRet));
		return FALSE;
	}

	pAddr->sin_family = AF_INET;
	pAddr->sin_port = htons(0);
	pAddr->sin_addr  = ((SOCKADDR_IN*)outAddrInfo->ai_addr)->sin_addr;

	if(NULL != outAddrInfo->ai_canonname &&
		nSize > strlen(outAddrInfo->ai_canonname) ) {
		memcpy( pHostName, outAddrInfo->ai_canonname, 
			strlen(outAddrInfo->ai_canonname)+1);
	}

	freeaddrinfo(outAddrInfo);

	return TRUE;
	
}


/////////////////////////////////////////////////////////////////////


void main(int argc, char* argv[]) 
{
	WSADATA wsaData;
	SOCKET s;
	SOCKADDR_IN AddrFrom, TargetAddr;
	char szBuff[sizeof(ICMP_HDR) + ICMP_DATASIZE], szBuffRecv[256];
	char* pHdrTop = NULL, szHostName[64] = "";
	INT nRecv, nRet, i;
	PIP_HEADER ip_hdr;
	PICMP_HDR icmp_header;
	struct timeval Timeout;
	fd_set readfds;
	DWORD dwTimeTaken;

	USHORT cs;

	// Parsing Command.
	if(2 != argc) {
		ShowUsage();
		return;
	}

	// Startup WinSock
	ZeroMemory(&wsaData, sizeof(WSADATA));
	WSAStartup(MAKEWORD(2, 2), &wsaData);

	if(!ResolveAddress(argv[1], &TargetAddr, szHostName, 64)) {
		return;
	}

	// Startup Message.
	printf("Pinging %s [%s] with %d bytes of data:\n\n",
		szHostName, inet_ntoa(TargetAddr.sin_addr), ICMP_DATASIZE);

	for( i=0; i<4; i++) {

		//
		// Creating a socket.
		//

		s = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);

		if( INVALID_SOCKET == s ) {
			printf("socket Failed. (%d)\n", WSAGetLastError());
			return;
		}

		// Preparing for after use.
		readfds.fd_count = 1;
		readfds.fd_array[0] = s;
		Timeout.tv_sec = 5; // timeout 5 sec.
		Timeout.tv_usec  =0;

		//
		// Build a ICMP Echo Packet.
		//
		
		ZeroMemory(szBuff, sizeof(szBuff));
		BuildICMPEchoPacket(szBuff);

		//
		// Send the Echo Packet.
		//
		
		nRet = sendto( s, szBuff, sizeof(ICMP_HDR) + ICMP_DATASIZE, 
			0, (SOCKADDR*) &TargetAddr, sizeof(TargetAddr));

		if( SOCKET_ERROR == nRet ) {
			printf("sendto Failed. (%d)\n", WSAGetLastError());
			break;
		}

		//
		// Waiting for receiving datagram.
		//
		
		nRet = select(1, &readfds, NULL, NULL, &Timeout);

		if(nRet) {

			//
			// Receive a data.
			//
			
			nRecv = sizeof(AddrFrom);
			nRet = recvfrom( s, szBuffRecv, sizeof(szBuffRecv), 0, (SOCKADDR*)&AddrFrom, &nRecv);

			if( 0 == nRet || SOCKET_ERROR == nRet ) {
				printf("recvfrom Failed. (ret, gle) = (%d, %d)\n", nRet, WSAGetLastError());
				break;
			}
			
			dwTimeTaken = GetTickCount();

			//
			// Preparing for reading headers.
			//
			
			ip_hdr = (PIP_HEADER) szBuffRecv;

			pHdrTop = szBuffRecv;
			pHdrTop += ((int) ip_hdr->ip_hl) * 4;
			icmp_header = (PICMP_HDR) pHdrTop;

			dwTimeTaken -= (DWORD) icmp_header->icmp_timestamp;

			cs = CheckSum((USHORT*)icmp_header, sizeof(ICMP_HDR)+ICMP_DATASIZE);
			
			//
			// Print Status.
			//
			if(0 == cs) {
				printf("Reply from %s: bytes=%d time=%dms TTL=%d\n", 
					inet_ntoa(ip_hdr->ip_src), 
					nRet - ((int) ip_hdr->ip_hl) * 4 - sizeof(ICMP_HDR),
					dwTimeTaken,
					ip_hdr->ip_ttl);

				PrintHexDump(sizeof(ICMP_HDR), (PBYTE) icmp_header);
				
			}
			else {
				printf("Data received corrupted.\n");
			}

		}
		else {

			//
			// Error Handling
			//
			
			if(0 == nRet) { // Timed out.
				printf("Request timed out.\n");
			}
			else { // Other Errors
				printf("select Failed. %d\n", WSAGetLastError());
			}
		}
				
		//
		// Close the socket.
		//
		
		closesocket(s);

		//
		// Just a moment, please.
		//
		
		Sleep(1000);
		
	}
	
	//
	WSACleanup();

}
